#!/usr/bin/python3
import argparse
import configparser
import logging

import signal
import threading
import socketserver
import time
import json
import hashlib
import importlib.util
import os
import fnmatch
import re

from datetime import datetime

from http.server import BaseHTTPRequestHandler, HTTPServer
from http import HTTPStatus
import http.client
from urllib.parse import urlparse, parse_qs
import cgi

shared_memory = dict()

class EsiaHTTPHandler(BaseHTTPRequestHandler):
    def log_request(self, code='-', size='-'):
        #Do nothing
        return

    def log_error(self, format, *args):
        #Do nothing
        return

    def getGETParams(self):
        logger.debug('Parsing HTTP GET Request parameters')
        http_keys = ['key','debug', 'action', 'proto', 'hostaddr', 'user', 'authentication','command']
        url = urlparse(self.path)
        get_params = parse_qs(url.query)

        http_params = dict()
        for key in http_keys:
            if key in get_params:
                http_params[key] = get_params[key][0]
            else:
                http_params[key] = None
        return http_params

    def getPOSTParams(self):
        logger.debug('Parsing HTTP POST Request parameters')
        form = cgi.FieldStorage(
            fp=self.rfile,
            headers=self.headers,
            environ={'REQUEST_METHOD': 'POST'}
        )

        http_keys = ['key','debug', 'action', 'proto', 'hostaddr', 'user', 'authentication','command']
        http_params = dict()
        for key in http_keys:
            http_params[key] = form.getfirst(key, None)
        return http_params

    def missingRequiredParams(self,http_params,needed_keys):
        missingRequired = []
        for key in needed_keys:
            if http_params[key] == None:
                missingRequired.append(key)
        return missingRequired

    def request_string(self,post_params):
        requestline = self.requestline
        requestline = re.sub("authentication=\S+", "authentication=<SHADOWED>", requestline)

        if 'authentication' in post_params:
            post_params['authentication'] = "<SHADOWED>"
            post_params = json.dumps(post_params)
        else:
            post_params = ''
        message = self.address_string() +' - '+ str(requestline) + ' ' + str(post_params)
        return message

    def openConnection(self,proto,http_params):
        conn = dict()
        conn['conn'] = proto.openConn(http_params['hostaddr'],http_params['user'],http_params['authentication'])
        conn['name'] = http_params['proto'] + ':' + http_params['hostaddr'] + ':' + http_params['user']
        dt = datetime.now()
        conn['uptime'] = datetime.timestamp(dt)
        conn['lock'] = threading.Lock()
        return conn

    def sendHttpResponse(self,code,data):
        try:
            self.send_response(code)
            self.send_header(b'Content-type',b'application/json')
            self.end_headers()
            message = json.dumps(data)
            self.wfile.write(message.encode())
        except (BrokenPipeError, IOError) as e:
            logger.error('Unable to send response code '+str(code.value)+' to client. '+str(e))

    def sendHttpError(self,code,action,description):
        response = dict()

        dt = datetime.now()
        response['date'] = datetime.timestamp(dt)
        response['thread'] = str(threading.current_thread())
        response['code'] = code
        response['message'] = http.client.responses[code]
        response['action'] = action
        response['more_info'] = description
        self.sendHttpResponse(code,response);
        logger.warning(description)

    def do_GET(self):
        http_params = self.getGETParams()
        try:
            code = self.do_request(http_params)
        except Exception as e:
            code = code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.sendHttpError(code,http_params['action'],str(e))
        request_str = self.request_string('')
        logger.info(str(code.value) + ' - ' + request_str)

    def do_POST(self):
        http_params = self.getPOSTParams()
        try:
            code = self.do_request(http_params)
        except Exception as e:
            code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.sendHttpError(code,http_params['action'],str(e))
        request_str = self.request_string(http_params)
        logger.info(str(code.value) + ' - ' + request_str)

    def do_request(self,http_params):
        global shared_memory

        shared_key = str(shared_memory['shared_key'])

        logger.debug('Checking presence of global HTTP parameters')
        missingParameters = self.missingRequiredParams(http_params,['key', 'proto'])
        if missingParameters:
            code = HTTPStatus.BAD_REQUEST
            self.sendHttpError(code,http_params['action'],"You forget global required parameters : "+str(missingParameters))
            return code

        logger.debug('Global HTTP parameters validation')
        if http_params['key'] != shared_key:
            code = HTTPStatus.FORBIDDEN
            self.sendHttpError(code,http_params['action'],"Incorrect shared key ('key' parameter) send")
            return code
        if http_params['proto'] in shared_memory['proto']:
            conn_list = shared_memory[http_params['proto']]
        else:
            code = HTTPStatus.BAD_REQUEST
            self.sendHttpError(code,http_params['action'],"Unknown protocol 'proto' : '"+ http_params['type'] +"' parameter")
            return code

		### ---------- RUN ACTION ---------- ###
        if http_params['action'] == 'run':
            logger.debug("STARTING action 'run'")
            logger.debug("Checking presence of 'run' action HTTP parameters")
            missingParameters = self.missingRequiredParams(http_params,['hostaddr', 'user', 'authentication', 'command'])
            if missingParameters:
                code = HTTPStatus.BAD_REQUEST
                self.sendHttpError(code,http_params['action'],"You forget action 'run' required parameters : "+str(missingParameters))
                return code

            logger.debug("Loading correct HTTP Module")
            try:
                script_dir = str(os.path.abspath( os.path.dirname( __file__ ) ))
                spec=importlib.util.spec_from_file_location("proto",script_dir+"/proto/"+http_params['proto']+".py")
                proto = importlib.util.module_from_spec(spec);
                spec.loader.exec_module(proto)
            except Exception as e:
                code = HTTPStatus.INTERNAL_SERVER_ERROR
                self.sendHttpError(code,http_params['action'],"Unable to load protocol module '"+http_params['proto']+"' : "+str(e))
                return code

            logger.debug("Generate connection hash")
            salt = str(shared_memory['startat']) + shared_key
            conn_string = salt + ':' + http_params['hostaddr'] + ':' + http_params['user'] + ':' + http_params['authentication']
            hash_object = hashlib.sha1(conn_string.encode())
            conn_token = hash_object.hexdigest()

            logger.debug("Getting connection")
            conn_list['lock'].acquire()
            if conn_token in conn_list['list']:
                logger.debug("Already connected")
                conn_current = conn_list['list'][conn_token]
                conn_list['lock'].release()
                conn_current['lock'].acquire()
                try:
                    logger.debug("Running code (old connection)")
                    data = proto.run(conn_current['conn'],http_params['command'])
                    code = HTTPStatus.OK
                    self.sendHttpResponse(code,data)
                    conn_current['lock'].release()
                    logger.debug("STOPPING action 'run'")
                    return code
                except Exception: #Si on reçoit une exception, on continue, sinon on quitte avec succès (ci-dessus)
                    logger.info("Connection lossed. Trying to reconnect...")
                    conn_current['lock'].release()
                    conn_list['lock'].acquire()

            logger.info("Openning connection")
            try:
                conn_current = self.openConnection(proto,http_params)
                conn_list['list'][conn_token] = conn_current
                conn_list['lock'].release()
            except Exception as e:
                conn_list['lock'].release()
                code = HTTPStatus.UNAUTHORIZED
                self.sendHttpError(code,http_params['action'],'Unable to establish connection: '+str(e))
                return code

            logger.debug("Running code (fresh connection)")
            conn_current['lock'].acquire()
            try:
                data = proto.run(conn_current['conn'],http_params['command'])
                code = HTTPStatus.OK
                self.sendHttpResponse(code,data)
                conn_current['lock'].release()
                logger.debug("STOPPING action 'run'")
                return code
            except Exception as e:
                code = HTTPStatus.NOT_FOUND
                self.sendHttpError(code,http_params['action'],'Unable to run command: '+str(e))
                conn_current['lock'].release()
                logger.debug("STOPPING action 'run'")
                return code

        ### ---------- GET ACTION ---------- ###
        elif http_params['action'] == 'get':
            logger.debug("STARTING action 'get'")
            conn_list['lock'].acquire()
            data = []
            for conn_token in conn_list['list']:
                data.append(conn_list['list'][conn_token]['name'])
            conn_list['lock'].release()

            code = HTTPStatus.OK
            self.sendHttpResponse(code,data)
            logger.debug("STOPPING action 'get'")
            return code
        ### ---------- UNKNOWN ACTION ---------- ###
        else:
            code = HTTPStatus.METHOD_NOT_ALLOWED
            self.sendHttpError(code,http_params['action'],"Unknown 'action' method")
            return code

class ThreadedHTTPServer(socketserver.ThreadingMixIn, HTTPServer):
	pass

process_next_loop = True
def handler(signum, frame):
    if( signum == signal.SIGTERM ):
        logger = logging.getLogger('main')
        logger.info('SIGTERM received')
        global process_next_loop
        process_next_loop = False

if __name__ == '__main__':
    ## urlib3 génére apparemment des erreurs HeaderParsingError via systemcl uniquement
    ## see : https://stackoverflow.com/questions/56781990/disable-headerparsingerror-appearing-from-python3-urllib3 to disable it
    urllib3_logger = logging.getLogger('urllib3')
    urllib3_logger.setLevel(logging.CRITICAL)

    argParser = argparse.ArgumentParser()
    argParser.add_argument("-c", "--conf",required=True ,help="use CONF as configuration file or directory")
    args = argParser.parse_args()

    config = configparser.ConfigParser()
    config.read(args.conf)

    pid = os.getpid()
    fmtstr = "%(asctime)s %(filename)s ["+str(pid)+ "] %(threadName)10s Line: %(lineno)4d %(funcName)15s - %(levelname)8s: %(message)s"
    datestr = "%Y/%m/%d %H:%M:%S"

    logging.basicConfig(
        filename=config.get('FILEPATH', 'logfile'),
        #level=getattr(logging, config.get('LEVEL', 'logfile').upper(), None),
    #    filemode="w",
        format=fmtstr,
        datefmt=datestr
    )

    logger = logging.getLogger('main')
    logger.setLevel(getattr(logging, config.get('LEVEL', 'logfile').upper(), None))
    ch = logging.StreamHandler()
    ch.setLevel(getattr(logging, config.get('LEVEL', 'stdout').upper(), None))
    ch.setFormatter(logging.Formatter(fmtstr))
    logger.addHandler(ch)

    logger.info('Daemon Started')

    dt = datetime.now()

    script_dir = str(os.path.abspath( os.path.dirname( __file__ ) ))

    shared_memory['startat'] = datetime.timestamp(dt)
    shared_memory['shared_key'] = config.get('LISTEN', 'key')
    shared_memory['proto'] = []
    for file in os.listdir( script_dir +'/proto'):
        if fnmatch.fnmatch(file, '*.py'):
            shared_memory['proto'].append(file)

    shared_memory['proto'] = list(map(lambda x: x.replace('.py', ''), shared_memory['proto']))
    logging.debug('Found protocols libs : ' + str(shared_memory['proto']))

    for proto in shared_memory['proto']:
        shared_memory[proto] = dict()
        shared_memory[proto]['lock'] = threading.Lock()
        shared_memory[proto]['list'] = dict()

    logger.debug('Protocols ready to use : ' + str(shared_memory['proto']))
    params = (config.get('LISTEN', 'address'),config.getint('LISTEN', 'port'))

    logger.debug('Starting HTTP Server : ' + str(params))
    server = ThreadedHTTPServer(params,EsiaHTTPHandler)
    server_thread = threading.Thread(target=server.serve_forever)
    server_thread.daemon = True
    server_thread.start()

    logger.debug("Server loop running in thread: " + str(server_thread.name))

    signal.signal(signal.SIGTERM, handler)
    while process_next_loop:
        time.sleep(1)

    server.shutdown()
    server.server_close()
    logger.info('Daemon Finished')
